﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.Linq;
using System.Data.OleDb;
using System.Linq.Expressions;
using System.Reflection;

// Reference and based on
//  -IQueryable Implementation
//     http://linqinaction.net/blogs/main/archive/2006/09/07/Linq_to_Amazon_source_code.aspx
//  -Desktop Search usage
//     http://blogs.msdn.com/coding4fun/archive/2007/01/05/1417884.aspx

namespace LinqToDesktopSearch
{
    /// <summary>
    /// クエリ計算クラス
    /// </summary>
    /// <typeparam name="TProj">クエリ後に射影される型。</typeparam>
    /// <typeparam name="TQuery">Desktop Search のクエリに用いられるエンティティ型。</typeparam>
    public class DesktopSearchQuery<TProj, TQuery> : IQueryable<TProj> where TQuery:new()
    {
        const string ConnectionStr = "Provider=Search.CollatorDSO;Extended Properties='Application=Windows';";
        Expression _expr;
        Type _dataType = typeof(TProj);
        string _name = null;

        public DesktopSearchQuery(Expression expression)
        {
            _expr = expression;
        }

        static Dictionary<string, FieldInfo> _dataFields
         = typeof(TQuery).GetFields().ToDictionary(f => f.Name);

        IEnumerable<TQuery> PerformeDesktopSearch()
        {
            if (_name == null)
            {
                yield break;
            }

            using (var conn = new OleDbConnection(ConnectionStr))
            {
                conn.Open();

                var names = from fileinfo in _dataFields.Values select string.Format("\"System.{0}\"", fileinfo.Name);
                string selectFieldsQuery = string.Join(", ", names.ToArray());

                var cmd = new OleDbCommand();
                cmd.Connection = conn;
                
                // TODO: 本来はここで _name 変数を OLEDB の世界でのエスケープ済み文字列に変換する必要がある
                cmd.CommandText = string.Format(
                    "SELECT {0} " +
                    "FROM systemindex WHERE CONTAINS(\"System.ItemNameDisplay\", '\"*{1}*\"')",
                    selectFieldsQuery, _name);

                using (var results = cmd.ExecuteReader())
                {
                    while (results.Read())
                    {
                        TQuery result = new TQuery();
                        for (int i = 0; i < results.FieldCount; i++)
                        {
                            if (!results.IsDBNull(i))
                            {
                                string fieldName = results.GetName(i);
                                if (fieldName.StartsWith("System.")) fieldName = fieldName.Substring(7);
                                object value = results.GetValue(i);
                                FieldInfo field = null;
                                if (_dataFields.TryGetValue(fieldName, out field))
                                {
                                    field.SetValue(result, value);
                                }
                            }
                        }
                        yield return result;
                    }
                }
            }
        }

        void ProcessExpression(Expression expression)
        {
            if (expression is MethodCallExpression)
            {
                ProcessMethodCall(expression as MethodCallExpression);
            }
            else if (expression is LambdaExpression)
            {
                ProcessExpression((expression as LambdaExpression).Body);
            }
            else if (expression.NodeType == ExpressionType.Quote)
            {
                ProcessExpression((expression as UnaryExpression).Operand);
            }
        }

        void ProcessMethodCall(MethodCallExpression expression)
        {
            switch (expression.Method.Name)
            {
                case "Contains":
                    if (expression.Object.NodeType == ExpressionType.MemberAccess)
                    {
                        var memberExpr = expression.Object as MemberExpression;
                        if (memberExpr.Member.DeclaringType == typeof(TQuery) &&
                            memberExpr.Member.Name == "ItemNameDisplay")
                        {
                            // file.Contains( _e ); とすると
                            // _e := expression.Arguments[0] である
                            // ラムダ式の実行時コンパイルを経由して _e を即時評価する

                            var _e = Expression.Lambda<Func<string>>(expression.Arguments[0]);
                            _name = _e.Compile()();
                        }
                    }
                    break;
            }
        }

        #region IQueryable<T> Members

        public IQueryable<TElement> CreateQuery<TElement>(Expression expression)
        {
            return new DesktopSearchQuery<TElement, TQuery>(expression);
        }

        public TResult Execute<TResult>(Expression expression)
        {
            throw new NotImplementedException();
        }

        #endregion

        #region IEnumerable<T> Members

        public IEnumerator<TProj> GetEnumerator()
        {
            var methodCall = _expr as MethodCallExpression;
            if (methodCall == null)
            {
                throw new Exception();
            }

            Expression<Func<TQuery, TProj>> proj = null;

            if (methodCall.Method.Name == "Select")
            {
                if (methodCall.Arguments[0] is MethodCallExpression &&
                    (methodCall.Arguments[0] as MethodCallExpression).Method.Name == "Where")
                {
                    var unaryExpr = methodCall.Arguments[1] as UnaryExpression;
                    proj = unaryExpr.Operand as Expression<Func<TQuery, TProj>>;
                    if (proj == null)
                    {
                        throw new Exception("This query is supported");
                    }
                    methodCall = methodCall.Arguments[0] as MethodCallExpression;
                }
            }

            if (methodCall.Method.Name != "Where")
            {
                throw new Exception("this type of query is not supported.");
            }

            foreach (var arg in methodCall.Arguments)
            {
                ProcessExpression(arg);
            }

            IEnumerable<TQuery> files = PerformeDesktopSearch();

            return (proj == null) ?
                files.GetEnumerator() as IEnumerator<TProj> :
                files.Select<TQuery, TProj>(proj.Compile()).GetEnumerator();
        }

        #endregion

        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion

        #region IQueryable Members

        public IQueryable CreateQuery(Expression expression)
        {
            throw new NotImplementedException();
        }

        public Type ElementType
        {
            get { return _dataType; }
        }

        public object Execute(Expression expression)
        {
            throw new NotImplementedException();
        }

        public Expression Expression
        {
            get { return _expr; }
        }

        #endregion
    }

    public class DesktopSearch<TQuery> : IQueryable<TQuery> where TQuery:new()
    {
        #region IQueryable<FileInfo> Members

        public IQueryable<TElement> CreateQuery<TElement>(Expression expression)
        {
            if (typeof(TElement) != typeof(TQuery))
            {
                throw new Exception("TElement should be " + typeof(TQuery).ToString() );
            }
            return new DesktopSearchQuery<TElement, TQuery>(expression) as IQueryable<TElement>;
        }

        public TResult Execute<TResult>(Expression expression)
        {
            throw new NotImplementedException();
        }

        #endregion

        #region IEnumerable<TQuery> Members

        public IEnumerator<TQuery> GetEnumerator()
        {
            throw new NotImplementedException();
        }

        #endregion

        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            throw new NotImplementedException();
        }

        #endregion

        #region IQueryable Members

        public IQueryable CreateQuery(Expression expression)
        {
            return CreateQuery<TQuery>(expression);
        }

        public Type ElementType
        {
            get { return typeof(TQuery); }
        }

        public object Execute(Expression expression)
        {
            return Execute<TQuery>(expression);
        }

        public Expression Expression
        {
            get { return Expression.Constant(this); }
        }

        #endregion
    }
}
